from asyncio import Event, sleep
from ipaddress import IPv4Address
from random import choice, randint

from rich import print

from .client_base import ClientBase
from .opcodes import RecvOps, SendOps
from .packet import iPacket, oPacket, packet_handler
from .packet_helper import Helpers
from .tools import big_fuckin_mac


# NAMES = []


class Client(ClientBase):
    def __init__(
        self,
        loop=None,
        names=None,
        username=None,
        password=None,
        proxy=None,
        pproxy=None,
    ):
        super().__init__(_loop=loop, proxy=proxy, pproxy=pproxy)
        self._username = username
        self._password = password

        self._worlds = {}
        self._world_status = 0
        self._world_id = 0
        self._channel_count = 0
        self._channel_id = 0
        self._characters = []
        self._character_id = 0
        self._char_selected = None
        self._NAMES = names

        self._char_name = ""

    async def process_events(self):
        while self._loop.is_running():
            action = await self._action_queue.get()
            match action:
                case SendOps.LOGIN_PASSWORD:
                    await self.do_login()
                case SendOps.CHARLIST_REQUEST:
                    if self._world_status < 2:
                        await self.request_char_list()
                case SendOps.PLAYER_LOGGEDIN:
                    await self.migrate_in()
                case _:
                    ...

            self._action_queue.task_done()
            await sleep(1)


    @packet_handler(RecvOps.PING)
    async def pong(self, ipkt: iPacket):
        opkt = oPacket(SendOps.PONG)
        await self.send_packet(opkt)

    @packet_handler(RecvOps.LOGIN_STATUS)
    async def login_first(self, ipkt: iPacket):
        r1 = ipkt.decode_byte()
        if r1 != 0:
            print(f"Login Response code: {r1}")
            return

        ipkt.decode_byte()
        ipkt.decode_int()
        self._account_id = ipkt.decode_int()
        gender = ipkt.decode_byte()
        gm_1 = ipkt.decode_byte()
        ipkt.decode_short()
        gm_2 = ipkt.decode_byte()
        self._account_name = ipkt.decode_string()
        new_account = ipkt.decode_byte()
        shadow_ban = ipkt.decode_byte()
        shadow_ban_duration = ipkt.decode_long()
        ipkt.decode_byte()
        ipkt.decode_long()
        ipkt.decode_int()
        self._pin_enabled = ipkt.decode_byte() == 0
        self._pic_enabled = (
            ret == 1
            if (ret := ipkt.decode_byte()) not in [0, 2]
            else None
            if ret == 0
            else False
        )
        ipkt.decode_long()

        print(
            f"Account ID: {self._account_id} | Account Name: {self._account_name} | PIN: {self._pin_enabled} | PIC: {self._pic_enabled} | Shadow Ban: {shadow_ban} "
        )

        await self.send_packet(oPacket(SendOps.SERVERLIST_REQUEST))

    @packet_handler(RecvOps.LOGIN_SECOND)
    async def login_second(self, ipkt: iPacket):
        self._can_login.set()

    @packet_handler(RecvOps.SERVERLIST)
    async def serverlist(self, ipkt: iPacket):
        if (world_id := ipkt.decode_byte()) == 0xFF:
            await self.send_packet(oPacket(SendOps.SERVERSTATUS_REQUEST))
            await self.request_char_list()
            return

        world_name = ipkt.decode_string()
        world_flag = ipkt.decode_byte()
        event_message = ipkt.decode_string()

        self._worlds[world_id] = {
            "name": world_name,
            "flag": world_flag,
            "message": event_message,
            "channels": {},
            "load": 0,
        }

        ipkt.decode_short()
        ipkt.decode_short()
        ipkt.decode_byte()

        self._channel_count = ipkt.decode_byte()

        for _ in range(self._channel_count):
            ch_name = ipkt.decode_string()
            load = ipkt.decode_int()
            w_id = ipkt.decode_byte()
            ch_id = ipkt.decode_short()
            self._worlds[w_id]["load"] += load
            self._worlds[w_id]["channels"][ch_id] = {}
            self._worlds[w_id]["channels"][ch_id]["load"] = load
            self._worlds[w_id]["channels"][ch_id]["name"] = ch_name

        ipkt.decode_short()
        ipkt.decode_int()

    @packet_handler(RecvOps.ENABLE_RECOMMENDED)
    async def enable_recommended(self, ipkt: iPacket):
        most_popular = ipkt.decode_int()

    @packet_handler(RecvOps.SEND_RECOMMENDED)
    async def recv_recommended(self, ipkt: iPacket):
        num_ballons = ipkt.decode_byte()
        world = ipkt.decode_int()
        message = ipkt.decode_string()

    @packet_handler(RecvOps.SERVERSTATUS)
    async def server_status(self, ipkt: iPacket):
        self._world_status = ipkt.decode_byte()

    @packet_handler(RecvOps.CHARLIST)
    async def char_list(self, ipkt: iPacket):
        self._characters = Helpers.character_entries(ipkt)
        await self._can_login.wait()

        if not self._characters:  # create char
            await self.check_name()
            return
        else:  # login to char
            self._char_selected = self._characters[0]
            self._character_id = self._char_selected.idx

        await self.select_character()

    @packet_handler(RecvOps.ADD_NEW_CHAR_ENTRY)
    async def add_new_char(self, ipkt: iPacket):
        success = ipkt.decode_byte() == 0
        if not success:
            return

        self._char_selected = Helpers.character_entry(ipkt)
        self._characters.append(self._char_selected)
        self._character_id = self._char_selected.idx

        await self.select_character()

    @packet_handler(RecvOps.CHAR_NAME_RESPONSE)
    async def char_name_resp(self, ipkt: iPacket):
        name = ipkt.decode_string()
        taken = ipkt.decode_byte() == 1
        if taken:
            await self.check_name()
        else:
            self._char_name = name
            await self.create_char()

    @packet_handler(RecvOps.SERVER_IP)
    async def server_ip(self, ipkt: iPacket):
        # ipkt.decode_short()
        # ip = ipkt.decode_int()
        # self._conn_port = ipkt.decode_short()
        # # self._client_id = ipkt.decode_int()
        # print(self._conn_port, self._migrating)
        # self._migrating.set()
        ...

    async def do_login(self):
        opkt = oPacket(SendOps.LOGIN_PASSWORD)
        opkt.encode_string(self._username)
        opkt.encode_string(self._password)

        await self.send_packet(opkt)

    async def request_char_list(self):
        opkt = oPacket(SendOps.CHARLIST_REQUEST)
        opkt.encode_byte(2)
        self._channel_id = randint(0, self._channel_count - 1) + 1
        opkt.encode_byte(self._world_id)
        opkt.encode_short(self._channel_id)

        await self.send_packet(opkt)

    async def check_name(self):
        opkt = oPacket(SendOps.CHECK_CHAR_NAME)
        opkt.encode_string(choice(self._NAMES))
        await self.send_packet(opkt)

    async def create_char(self):
        opkt = oPacket(SendOps.CREATE_CHAR)
        opkt.encode_string(self._char_name)
        job = 1
        opkt.encode_int(job)  # job
        opkt.encode_short(0)  # dual blade
        opkt.encode_byte(0)  # gender
        opkt.encode_byte(1)  # skin color
        opkt.encode_byte(8)
        faces = [20000, 20001, 20002]
        hairs = [30030, 30020, 30000]
        hair_colors = [0, 2, 3, 7]
        eye_colors = [0, 1, 2, 3]
        shirts = [1040002, 1040006, 1040010]
        ds_marks = [1012277, 1012278, 1012279, 1012280]
        shorts = [1060002, 1060006]
        shoes = [1072001, 1072005, 1072037, 1072038]
        weapons = [1302000, 1322005, 1312004]
        opkt.encode_int(choice(faces))  # face
        opkt.encode_int(choice(hairs))  # hair
        if job not in [5, 6]:
            opkt.encode_int(choice(hair_colors))  # hair color
            opkt.encode_int(choice(eye_colors))  # eye color
        if job == 6:
            opkt.encode_int(choice(ds_marks))

        opkt.encode_int(choice(shirts))
        opkt.encode_int(choice(shorts))
        opkt.encode_int(choice(shoes))
        opkt.encode_int(choice(weapons))
        await self.send_packet(opkt)

    async def select_character(self):
        opkt = oPacket(SendOps.CHAR_SELECT_NO_PIC)
        opkt.encode_int(self._character_id)
        mac = big_fuckin_mac()
        mac2 = f'{mac.strip("-")}_8FE7F2CA'
        opkt.encode_string(mac)
        opkt.encode_string(mac2)
        await self.send_packet(opkt)

    async def migrate_in(self):
        opkt = oPacket(SendOps.PLAYER_LOGGEDIN)
        opkt.encode_int(self._character_id)
        opkt.skip(6)
        opkt.encode_int(randint(0, 2147483647))
        opkt.encode_int(0)
        opkt.encode_short(randint(0, 32767))
        opkt.encode_long(self._client_id)
        await self.send_packet(opkt)
