from asyncio import (Event, Lock, Queue, Task, TimeoutError, create_task,
                     get_event_loop, sleep)
from datetime import datetime
from socket import AF_INET, IPPROTO_TCP, SOCK_STREAM, TCP_NODELAY, socket
from logging import log, basicConfig
basicConfig(filename="clb.log", filemode="a", level=10)

from socks import socksocket, SOCKS5, create_connection

from random import choice
try:
    from python_socks import ProxyConnectionError
    from python_socks.async_.asyncio._connect import connect_tcp
    from python_socks.async_.asyncio._proxy import Socks5Proxy
except ImportError:
    print("python_socks is required to enable proxied connections, all connections made will be unproxied.") # type: ignore

try:
    from rich import print
except ImportError:
    ...

from re import compile

from .crypto import MapleAes, MapleIV, decrypt_transform, encrypt_transform
from .opcodes import RecvOps, SendOps
from .packet import PacketHandler, iPacket
from .proxies import public, private

PROXY_RE = compile(r"(?=^(?:([^:]+):([^@]+)@)?([^:]+):([^$]+)$)")
JUNK = set()
LOGIN_PORT = 8485

class ClientBase:

    def __init__(self, _loop=None, proxy=None, pproxy=None):
        self._loop = _loop or get_event_loop()

        self._proxy_url = proxy or ""
        match = PROXY_RE.match(self._proxy_url)

        # if match:
        if len(match.groups()) > 2:
            self._prox_user, self._prox_pass, self._prox_ip, self._prox_port = match.groups()
            self._prox_port = int(self._prox_port)
        else:
            self._prox_ip, self._prox_port = match.groups()
            self._prox_port = int(self._prox_port)
            self._prox_user = None
            self._prox_pass = None

        print(f"User: {self._prox_user} | Password: {self._prox_pass} | IP: {self._prox_ip} | Port: {self._prox_port}")

        self._tasks = []
        self._packet_handlers = []
        self._can_login: Event = Event()
        self._migrating: Event = Event()
        self._change_channel: Event = Event()
        self._migrated: Event = Event()
        self._wait_sock_connect = Event()
        self._lock: Lock = Lock()
        self._action_queue: Queue = Queue()
        self._events_task: Task

        self._conn_port = 0
        self._sock: socket

        self._version = None
        self._sub_version = None
        self._locale = None

        self._recv_iv: MapleIV = MapleIV()
        self._send_iv: MapleIV = MapleIV()

        self._buff = bytearray()
        self._prev_buff = bytearray()

        self.add_packet_handlers()

    async def _connect_sock(self):
        # self._sock = socksocket(family=AF_INET, type=SOCK_STREAM, proto=IPPROTO_TCP)
        # self._sock.setblocking(False)
        # self._sock.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)
        # self._sock.setproxy(proxy_type=SOCKS5, addr=self._prox_ip, port=self._prox_port, username=self._prox_user, password=self._prox_pass)
        # await self._loop.sock_connect(self._sock, (self._host, self._conn_port))

        # if hasattr(self, "_sock"):
        #     self._sock.close()
        #     del self._sock

        # while not hasattr(self, "_sock"):
        if hasattr(self, "_prox_user"):
            self._conn = Socks5Proxy(proxy_host=self._prox_ip, proxy_port=self._prox_port, username=self._prox_user, password=self._prox_pass, loop=self._loop)
            # print(self._conn_port)
        else:
            self._conn = Socks5Proxy(proxy_host=self._prox_ip, proxy_port=self._prox_port, loop=self._loop)

        try:
            self._sock = await self._conn.connect(self._host, self._conn_port, timeout=10.0)
        except Exception as e:
            print(f"{e} eaaaaaaaaaaaaaa")
            # try:
            #     self._sock = await self._conn.connect(self._host, self._conn_port)
            # except ProxyConnectionError:
            #     print(f"{ProxyConnectionError.__name__} | Unable connect to proxy socket")
            #     JUNK.add(self._proxy_url)
            #     # del self._events_task
            #     # return False
            # except TimeoutError:
            #     print(f"{TimeoutError.__name__} | Unable to connect to proxy")
            #     JUNK.add(self._proxy_url)
            #     # del self._events_task
            #     # return False
            # except Exception as e:
            #     print(f"Unhandled Exception | {e.__class__.__name__}")
            #     # continue
            #     # del self._events_task

            # if not self._migrating.is_set():
            #     if not hasattr(self, "_sock"):
            #         prox = choice(public).split(":")
            #         self._prox_ip = prox[0]
            #         self._prox_port = prox[1]
            #         continue


        self._wait_sock_connect.set()

    async def start(self):
        self._loop.create_task(self.process_events())

        self._host = "51.222.56.169"
        self._conn_port = LOGIN_PORT

        self._loop.create_task(self._sock_recv())

    async def send_packet(self, opacket, raw=False):
        async with self._lock:
            opkt = bytearray(opacket.getvalue())
            length = len(opkt)

            buf = bytearray(length + 4)
            buf[0:4] = self._send_iv.get_header(self._version, length)
            opkt = bytearray(encrypt_transform(opkt))
            MapleAes.transform(opkt, self._send_iv)
            buf[4 : length + 4] = opkt

            # print(f"[{datetime.now().strftime('%d %H:%M:%S.%f')}] [SEND] : {SendOps(opacket.op_code).name} | {opacket.to_string()}")

            await self._wait_sock_connect.wait()

            # if not self._sock:
            #     return

            await self._loop.sock_sendall(self._sock, buf)

    def schedule_action(self, op_code, delay: None | float = 1.0):
        async def schd():
            if delay:
                await sleep(delay)

            await self._action_queue.put(op_code)

        self._loop.create_task(schd())

    async def _sock_recv(self):
        await self._connect_sock()

        while self._loop.is_running():
            await self._wait_sock_connect.wait()
            try:
                while hasattr(self, "_sock"):
                    await sleep(1)
                    self._buff.extend(bytearray((await self._loop.sock_recv(self._sock, 8192))))

                    if self._buff == bytearray(b"\x00\x00\x00\x00\x00\x00") and len(self._buff) == 6:
                        break

                    if not self._send_iv and self._buff:
                        begin_packet = iPacket(self._buff)
                        op_code = 0x0E
                        # print(f"[{datetime.now().strftime('%d %H:%M:%S.%f')}] [Client Hello]")

                        self._version = begin_packet.decode_short()
                        self._sub_version = int(begin_packet.decode_string())
                        self._send_iv(begin_packet.decode_int())
                        self._recv_iv(begin_packet.decode_int())
                        self._locale = begin_packet.decode_byte()

                        self._buff = bytearray()

                        if self._conn_port != LOGIN_PORT and self._migrating.is_set():
                            print(f"[{datetime.now().strftime('%d %H:%M:%S.%f')}] [Client Hello] [Player Logged In]")
                            log(20, msg=f"IP - {self._prox_ip} Port - {self._prox_port} LoginPort - {self._conn_port}")
                            # await sleep(2.0)
                            # self._loop.create_task(self.migrate_in())
                            self.schedule_action(SendOps.PLAYER_LOGGEDIN, 2.0)
                            self._migrated.set()
                            self._migrating.clear()
                        else:
                            print(f"[{datetime.now().strftime('%d %H:%M:%S.%f')}] [Client Hello] [Client Begin]")
                            # await sleep(2.0)
                            # self._loop.create_task(self.do_login())
                            self.schedule_action(SendOps.LOGIN_PASSWORD, 2.0)

                    else:
                        migrate = False
                        while len(self._buff) > 0:
                            try:
                                length = MapleAes.get_length(self._buff[0:4])

                                buf = bytearray(bytearray(self._buff)[4 : length + 4])

                                buf = bytearray(MapleAes.transform(buf, self._recv_iv))
                                buf = decrypt_transform(buf)
                                packet = iPacket(buf)

                                migrate = await self.handle_packet(packet)
                                if migrate:
                                    break

                                if len(self._buff) - (length + 4) > 0:
                                    self._buff = self._buff[length + 4 :]

                                else:
                                    self._buff = bytearray()

                                self._prev_buff = self._buff
                            except Exception as e:
                                print(f"buffer error: {e}")
                                break

                        if migrate:
                            break

                else:
                    print(f"Sock got terminated")

            except Exception as e:
                print(f"Exception within sock_recv | {e}")

            finally:
                # self._sock.close()
                print(f"[{datetime.now().strftime('%d %H:%M:%S.%f')}] closing socket")

            print(self._conn_port)
            if self._conn_port != LOGIN_PORT:
                if not self._migrating.is_set() and self._change_channel.is_set():
                    self._migrating.set()
                    self._change_channel.clear()
                elif self._migrated.is_set() and not self._change_channel.is_set() and self._migrating.is_set():
                    self._migrating.clear()
                    self._migrated.clear()

            self._recv_iv.reset()
            self._send_iv.reset()
            self._buff = bytearray()
            self._prev_buff = bytearray()
            self._wait_sock_connect.clear()
            await self._connect_sock()

    async def handle_packet(self, packet):
        coro = None

        for pkt_hndl in self._packet_handlers:
            if pkt_hndl.op_code == packet.op_code:
                coro = pkt_hndl.callback
                break
        else:
            # print(
            #     f"[{datetime.now().strftime('%d %H:%M:%S.%f')}] [RECV] [UNHANDLED] : {RecvOps(packet.op_code).name} | {packet.to_string()}"
            # )
            return False

        if not coro:
            return False


        # print(f"[{datetime.now().strftime('%d %H:%M:%S.%f')}] [RECV] : {RecvOps(packet.op_code).name}")# | {packet.to_string()}")

        if packet.op_code == RecvOps.SERVER_IP:
            self._migrating.set()
            packet.decode_short()
            packet.decode_int()
            self._conn_port = packet.decode_short()
            self._client_id = packet.decode_int()
            return True

        else:
            create_task(coro(self, packet))
            return False


    def add_packet_handlers(self):
        import inspect

        members = inspect.getmembers(self)
        for _, member in members:

            if (
                isinstance(member, PacketHandler)
                and member not in self._packet_handlers
            ):
                self._packet_handlers.append(member)

    async def process_events(self):
        raise NotImplementedError

    async def migrate_in(self):
        raise NotImplementedError

    async def do_login(self):
        raise NotImplementedError